use simcore_rs::{Process, Time, SimContext, Facility, RandomVar, simulation};
use rand::{distributions::Uniform, rngs::SmallRng, SeedableRng, Rng};

// helper constants
const STOP_TIME: Time = 480000.0;
const SEED_A   :  u64 = 100000;
const SEED_S   :  u64 = 200000;

/// Customer process with access to the barber and a random number generator.
struct Customer<'c> { joe: &'c Facility, rv: &'c RandomVar, rng: SmallRng }

impl<'c> Customer<'c> {
	pub async fn actions(mut self, sim: SimContext<'c>) {
		// access the barber and record the time for the report
		let time = sim.now();
		self.joe.seize().await;
		self.rv.tabulate(sim.now() - time);
		
		// spend time
		sim.advance(self.rng.gen_range(12.0, 18.0)).await;
		// release the barber
		self.joe.release();
	}
}

async fn sim_main<'m>(sim: SimContext<'m>, joe: &'m Facility, rv: &'m RandomVar) {
	// pseudo random number generators referenced from within the main process
	let mut rng_a = SmallRng::seed_from_u64(SEED_A);
	let mut rng_s = SmallRng::seed_from_u64(SEED_S);
	
	// activate a process to generate the customers
	sim.activate(async move {
		let dist = Uniform::new(12.0, 24.0);
		loop {
			sim.advance(rng_a.sample(dist)).await;
			if sim.now() >= STOP_TIME { return; }
			sim.activate(Customer {
				joe: joe.clone(),
				rv: rv.clone(),
				rng: SmallRng::from_seed(rng_s.gen())
			}.actions(sim));
		}
	});
	
	// wait until the store closes
	sim.advance(STOP_TIME).await;
	
	// finish processing the queue (no more customers arrive)
	joe.seize().await;
}

fn main() {
	// the lifetimes are automatically extended until the end of the scope
	let joe = &Facility::new();
	let rv = &RandomVar::new();
	
	// the main process
	simulation(|sim| Process::new(sim_main(sim, joe, rv)));
	
	println!("Stats: {:.3}", rv);
}
